﻿using System;
using System.Collections.Generic;
using System.Configuration;
using System.Linq;
using System.Security.Policy;
using System.Text;
using System.Threading.Tasks;
using HeuristicLab.Common;
using HeuristicLab.Problems.GrammaticalOptimization;

namespace HeuristicLab.Algorithms.Bandits.GrammarPolicies {
  public class EpsGreedyPolicy : GrammarPolicy {

    private readonly HashSet<string> done;
    private readonly double eps;
    private readonly IGrammarPolicy greedyPolicy;
    private readonly IGrammarPolicy randomPolicy;

    public EpsGreedyPolicy(IProblem problem, bool useCanonicalRepresentation = false, double eps = 0.1)
      : base(problem, useCanonicalRepresentation) {
      this.eps = eps;
      this.done = new HashSet<string>();
      this.greedyPolicy = new GreedyPolicy(problem, useCanonicalRepresentation);
      this.randomPolicy = new RandomPolicy(problem, useCanonicalRepresentation);
    }

    public override bool TrySelect(Random random, ReadonlySequence curState, IEnumerable<ReadonlySequence> afterStates, out ReadonlySequence selectedState) {
      // only select states that are not yet done
      afterStates = afterStates.Where(a => !done.Contains(CanonicalState(a.ToString()))).ToArray();
      if (!afterStates.Any()) {
        // fail because all follow states have already been visited => also disable the current state
        done.Add(CanonicalState(curState.ToString()));
        selectedState = null;
        return false;
      }

      if (random.NextDouble() > eps) {
        return greedyPolicy.TrySelect(random, curState, afterStates, out selectedState);
      } else {
        return randomPolicy.TrySelect(random, curState, afterStates, out selectedState);
      }
    }

    public override void UpdateReward(IEnumerable<ReadonlySequence> stateTrajectory, double reward) {
      base.UpdateReward(stateTrajectory, reward);
      // the last state could be terminal
      var lastState = stateTrajectory.Last();
      if (lastState.IsTerminal) done.Add(CanonicalState(lastState.ToString()));

      // forward
      greedyPolicy.UpdateReward(stateTrajectory, reward);
      randomPolicy.UpdateReward(stateTrajectory, reward); // is ignored but for completeness
    }

    public override void Reset() {
      base.Reset();
      done.Clear();
      greedyPolicy.Reset();
      randomPolicy.Reset();
    }

    public double GetValue(ReadonlySequence state) {
      return greedyPolicy.GetValue(state);
    }
  }
}
