﻿using System;
using System.Collections.Generic;
using System.Configuration;
using System.Linq;
using System.Security.Policy;
using System.Text;
using System.Threading.Tasks;
using HeuristicLab.Common;
using HeuristicLab.Problems.GrammaticalOptimization;

namespace HeuristicLab.Algorithms.Bandits.GrammarPolicies {
  public class UCTPolicy : GrammarPolicy {

    private readonly HashSet<string> done;
    private readonly double c;

    public UCTPolicy(IProblem problem, bool useCanonicalRepresentation = false, double c = 1.0)
      : base(problem, useCanonicalRepresentation) {
      this.c = c;
      this.done = new HashSet<string>();
    }

    public override bool TrySelect(Random random, ReadonlySequence curState, IEnumerable<ReadonlySequence> afterStates, out ReadonlySequence selectedState) {
      // only select states that are not yet done
      afterStates = afterStates.Where(a => !done.Contains(CanonicalState(a.ToString()))).ToArray();
      if (!afterStates.Any()) {
        // fail because all follow states have already been visited => also disable the current state
        done.Add(CanonicalState(curState.ToString()));
        selectedState = null;
        return false;
      }

      var totalTries = afterStates.Select(GetTries).Sum();
      var bestQ = double.NegativeInfinity;
      selectedState = null;
      foreach (var state in afterStates) {
        // try each state at least once
        if (GetTries(state) == 0) {
          selectedState = state;
          return true;
        }
        var q = AvgReward(state) + 2.0 * c * Math.Sqrt(Math.Log(totalTries) / GetTries(state));
        if (q > bestQ) {
          bestQ = q;
          selectedState = state;
        }
      }

      return true;
    }

    public override void UpdateReward(IEnumerable<ReadonlySequence> stateTrajectory, double reward) {
      base.UpdateReward(stateTrajectory, reward);
      // the last state could be terminal
      var lastState = stateTrajectory.Last();
      if (lastState.IsTerminal) done.Add(CanonicalState(lastState.ToString()));
    }

    public void Reset() {
      base.Reset();
      done.Clear();
    }
  }
}
