﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Resources;
using System.Runtime.InteropServices;
using System.Text;
using HeuristicLab.Algorithms.Bandits;
using HeuristicLab.Algorithms.Bandits.BanditPolicies;
using HeuristicLab.Algorithms.Bandits.GrammarPolicies;
using HeuristicLab.Common;
using HeuristicLab.Problems.GrammaticalOptimization;

namespace HeuristicLab.Algorithms.GrammaticalOptimization {
  // a search procedure that uses a policy to generate sentences and updates the policy (online RL)
  // 1) Start with phrase = sentence symbol of grammar
  // 2) Repeat
  //    a) generate derived phrases using left-canonical derivation and grammar rules
  //    b) keep only the phrases which are allowed (sentence length limit)
  //    c) if the set of phrases is empty restart with 1)
  //    d) otherwise use policy to select one of the possible derived phrases as active phrase
  //       the policy has the option to fail (for instance if all derived phrases are terminal and should not be visited again), in this case we restart at 1
  //    ... until phrase is terminal
  // 3) Collect reward and update policy (feedback: state of visited rewards from step 2)
  public class SequentialSearch {

    public event Action<string, double> FoundNewBestSolution;
    public event Action<string, double> SolutionEvaluated;

    private readonly int maxLen;
    private readonly IProblem problem;
    private readonly Random random;
    private readonly int randomTries;
    private readonly IGrammarPolicy behaviourPolicy;
    private readonly IGrammarPolicy greedyPolicy;
    private int maxSearchDepth;

    private double bestQuality;
    private string bestPhrase;
    private int tries;
    private readonly List<ReadonlySequence> stateChain;

    public SequentialSearch(IProblem problem, int maxLen, Random random, int randomTries, IGrammarPolicy behaviourPolicy) {
      this.maxLen = maxLen;
      this.problem = problem;
      this.random = random;
      this.randomTries = randomTries;
      this.behaviourPolicy = behaviourPolicy;
      this.greedyPolicy = new GreedyPolicy(problem, false);
      this.stateChain = new List<ReadonlySequence>();
      this.cache = new Dictionary<ReadonlySequence, ReadonlySequence[]>();
    }

    public void Run(int maxIterations) {
      bestQuality = double.MinValue;
      //InitPolicies(problem.Grammar);
      Reset();

      for (int i = 0; !Done() && i < maxIterations; i++) {
        var phrase = SampleSentence(problem.Grammar);
        // can fail on the last sentence
        if (phrase.IsTerminal) {
          var sentence = phrase.ToString();
          tries++;
          var quality = problem.Evaluate(sentence) / problem.BestKnownQuality(maxLen);
          Debug.Assert(quality >= 0 && quality <= 1.0);
          DistributeReward(quality);

          RaiseSolutionEvaluated(sentence, quality);

          if (quality > bestQuality) {
            bestQuality = quality;
            bestPhrase = sentence;
            RaiseFoundNewBestSolution(sentence, quality);
          }
        }
      }

      // clean up
      // Reset(); GC.Collect();
    }


    private ReadonlySequence SampleSentence(IGrammar grammar) {
      ReadonlySequence phrase;
      do {
        stateChain.Clear();
        phrase = new ReadonlySequence(grammar.SentenceSymbol);
        //var startPhrase = new Sequence("a*b+c*d+e*f+E");
      } while (!Done() && !TryCompleteSentence(grammar, ref phrase));
      return phrase;
    }

    private bool TryCompleteSentence(IGrammar g, ref ReadonlySequence phrase) {
      if (phrase.Length > maxLen) throw new ArgumentException();
      if (g.MinPhraseLength(phrase) > maxLen) throw new ArgumentException();
      var curDepth = 0;

      stateChain.Add(phrase);
      while (!phrase.IsTerminal) {

        //if (n.randomTries < randomTries) {
        //  n.randomTries++;
        //  treeDepth = Math.Max(treeDepth, curDepth);
        //  lastNode = n;
        //  return g.CompleteSentenceRandomly(random, phrase, maxLen);
        //} else {

        var newPhrases = GenerateFollowStates(g, phrase);

        // => select using bandit policy
        // failure means we simply restart
        if (!behaviourPolicy.TrySelect(random, phrase, newPhrases, out phrase)) {
          return false;
        }
        stateChain.Add(phrase);
        curDepth++;
      } // while

      maxSearchDepth = Math.Max(maxSearchDepth, curDepth);
      return true;
    }


    private readonly Dictionary<ReadonlySequence, ReadonlySequence[]> cache;
    private IEnumerable<ReadonlySequence> GenerateFollowStates(IGrammar g, ReadonlySequence phrase) {
      ReadonlySequence[] follow;
      if (!cache.TryGetValue(phrase, out follow)) {
        char nt = phrase.FirstNonTerminal;

        int maxLenOfReplacement = maxLen - (phrase.Length - 1);
        // replacing aAb with maxLen 4 means we can only use alternatives with a minPhraseLen <= 2
        Debug.Assert(maxLenOfReplacement > 0);

        var alts = g.GetAlternatives(nt).Where(alt => g.MinPhraseLength(alt) <= maxLenOfReplacement);

        follow = new ReadonlySequence[alts.Count()];
        int idx = 0;
        foreach (var alt in alts) {
          var newPhrase = new Sequence(phrase); // clone
          newPhrase.ReplaceAt(newPhrase.FirstNonTerminalIndex, 1, alt);
          follow[idx++] = new ReadonlySequence(newPhrase);
        }
        cache[phrase] = follow;
      }
      return follow;
    }

    private void DistributeReward(double reward) {
      behaviourPolicy.UpdateReward(stateChain, reward);
      greedyPolicy.UpdateReward(stateChain, reward);
    }


    private void Reset() {
      behaviourPolicy.Reset();
      greedyPolicy.Reset();
      maxSearchDepth = 0;
      bestQuality = 0.0;
      tries = 0;
      cache.Clear();
    }

    public bool Done() {
      var g = problem.Grammar;
      var startState = new ReadonlySequence(g.SentenceSymbol);
      var follow = GenerateFollowStates(g, startState);
      ReadonlySequence selectedState;
      return !behaviourPolicy.TrySelect(random, startState, follow, out selectedState);
    }

    #region introspection
    public void PrintStats() {
      Console.WriteLine("depth: {0,5} tries: {1,5} best phrase {2,50} bestQ {3:F3}", maxSearchDepth, tries, bestPhrase, bestQuality);

      // use greedy strategy to generate the currently prefered sentence
      var phrase = new ReadonlySequence(problem.Grammar.SentenceSymbol);
      var policy = behaviourPolicy;
      while (!phrase.IsTerminal) {
        Console.ForegroundColor = ConsoleColor.White;
        Console.WriteLine("{0,-30}", phrase);
        var newPhrases = GenerateFollowStates(problem.Grammar, phrase);
        if (!newPhrases.Any()) break;
        var values = newPhrases.Select(p => policy.GetValue(p));
        var maxValue = values.Max();
        if (maxValue == 0) maxValue = 1.0;

        // write phrases
        foreach (var p in newPhrases) {
          SetColorForValue(policy.GetValue(p) / maxValue);
          Console.Write(" {0,-4}", p.Subsequence(Math.Max(0, p.Length - 3), Math.Min(3, p.Length)));
        }
        Console.WriteLine();

        // write values
        foreach (var p in newPhrases) {
          SetColorForValue(policy.GetValue(p) / maxValue);
          Console.Write(" {0:F2}", policy.GetValue(p) * 10.0);
        }
        Console.WriteLine();

        // write tries
        foreach (var p in newPhrases) {
          SetColorForValue(policy.GetValue(p) / maxValue);
          Console.Write(" {0,4}", policy.GetTries(p));
        }
        Console.WriteLine();
        if (!policy.TrySelect(random, phrase, newPhrases, out phrase)) {
          break;
        }
      }

      Console.ForegroundColor = ConsoleColor.White;
      Console.WriteLine("-------------------");
    }

    private void SetColorForValue(double v) {
      Console.ForegroundColor = ConsoleEx.ColorForValue(v);
    }
    #endregion

    private void RaiseSolutionEvaluated(string sentence, double quality) {
      var handler = SolutionEvaluated;
      if (handler != null) handler(sentence, quality);
    }
    private void RaiseFoundNewBestSolution(string sentence, double quality) {
      var handler = FoundNewBestSolution;
      if (handler != null) handler(sentence, quality);
    }
  }
}
